/**
 * ECS root file
 * 
 * Contributors:
 * Brychan Dempsey - brychand@hotmail.com
 * 
 * References:
 * https://code.austinmorlan.com/austin/ecs - 'A simple C++ Entity Component System' - released under MIT licence
 *  
 */

import java.util.*;

public class ECS {
    // As mentioned by Austin Morlan, a queue is a better choice of data structure for the entity list
    // - it allows the first element to be efficiently popped from the list
    Queue<Integer> unusedEntities = new ArrayList<>();

    // As the entity subscribes (or does not) to a component, we can use a boolean state to represent the subscription
    // If we instead store the list of indicies the component is subscribed to, we would use significantly more 
    // memory space due to the components being defined as a 32-bit integer.
    // In a BitSet, each bit is a flag. Therefore we can represent subscription to the first 32 components in the
    // same space as one as a 32-bit int.
    //
    // This list therefore uses the index of the BitSet as its reference to the entity.
    List<BitSet> entityRegistrations = new ArrayList<>();

    // Each component can be assigned to every possible entity. There is a reasonably unlimited number of possible components;
    // so therefore we get an array of arrays containing all of a component's data allocations.
    // Access is in [componentType][entityNumber] format; getting the data from a subscribed entity is performed like so:
    // (<datatype>)componentDataArrays.get(ComponentIndex).get(Entity);
    // Unfortunately, given Java's constraints on dynamic typing (run-time type determination), the returned data must be
    // explicitly cast. Therefore, all components must implement the IComponent interface - which mandates the getType()
    // function

    // As discussed by Austin Morlan, this array must be packed, otherwise we waste time iterating over non-subscribed entities
    // This also means we cannot use the location implicitly - we must keep track of exactly which data instance associates with
    // which entity.
    // Therefore, we need a custom data structure that keeps reversable references to component instances and entities:
    interface IComponentArray{
        void entityDestroyed(int entity);
    }
    class ComponentArray<E> implements IComponentArray{
        List<E> componentArray = new ArrayList<>();
        Map<Integer, Integer> entityComponentDataMap = new HashMap<>();
        Map<Integer, Integer> componentDataEntityMap = new HashMap<>();

        public void entityDestroyed(int entity){
            Optional<Integer> pos = entityComponentDataMap.get(entity);
            if (pos.isEmpty()){
                removeData(entity);
            }
        }

        void removeData(int entity){
            if (!entityComponentDataMap.containsKey(entity)){
                System.err.println("Attempted to remove non-existent entity");
                return;
            }
            // Get the componentData index of the entity
            int removedComponentDataIndex = entityComponentDataMap.get(entity);
            // Replace the removed component with the last component in the array
            componentArray.set(removedComponentDataIndex, componentArray.get(componentArray.size() -1));
            // update the data positions in the map
            int lastEntity = componentDataEntityMap.get(componentArray.size()-1);
            entityComponentDataMap.replace(lastEntity, removedComponentDataIndex);
            componentDataEntityMap.replace(removedComponentDataIndex, lastEntity);
            // Finally, remomve the last elements
            entityComponentDataMap.remove(entity);
            componentDataEntityMap.remove(componentArray.size() -1);

            componentArray.remove(componentArray.size() -1);
        }

        void insertData(int entity, E component){
            Optional<Integer> pos = entityComponentDataMap.get(entity);
            if (!pos.isEmpty()){
                System.err.println("Entity is already subscribed to the component");
                return;
            }
            int index = componentArray.size();
            entityComponentDataMap.put(entity, index);
            componentDataEntityMap.put(index, entity);

            componentArray.add(component);
        }

        E getData(int entity){
            if (!entityComponentDataMap.containsKey(entity)){
                System.err.println("Attempted to retrieve non-existent data");
                return null;
            }

            return componentArray.get(entityComponentDataMap.get(entity));
        }
    }
    // The actual list of data arrays
    List<IComponentArray> componentDataArrays = new ArrayList<>();
    

    // Components are either primitive types or class/struct definitions, with an additional list associating
    // the entities the component is assigned to
    List<Object> components = new ArrayList<>();
    
    // Systems are user-defined functions that are called by the game engine, usually at a regular interval such 
    // as between every render frame. It must be able to find all instances of a specific component, in order
    // to update/read its data. A component  
    List<Object> systems = new ArrayList<>();


    int entityIndex = 0;
    int componentIndex = 0;
    int systemIndex = 0;

    /**
     * Takes the next element off the list of unassigned entities
     * @return the value of the new entity
     */
    Integer createEntity(){
        int newEntity = entities.remove(0);
        if (componentAssociations.size() <= newEntity){
            for (int i = componentAssociations.size(); i < newEntity+1 - componentAssociations.size(); i++) {
                componentAssociations.add(i, new ArrayList<>());
            }
        }
        componentAssociations.add(newEntity, element);
        return newEntity;
    }

    /**
     * Adds the entity back to the list of unassigned entities,
     * and empty all component associations
     * @param entity
     */
    void destroyEntity(int entity){
        entities.add(entity);
        componentAssociations.set(entity, new ArrayList<>());
    }

    void init(int maxEntities){
        for (int i = 0; i < maxEntities; i++) {
            entities.add(i);
        }
    }

}